﻿namespace DragAndDropList.Views.Behaviors
{
    using System;
    using System.Collections;
    using System.Windows;
    using System.Windows.Controls;
    using System.Windows.Input;

    public class ItemDragDropBehavior
    {
        #region IsEnabled 添付プロパティ
        public static readonly DependencyProperty IsEnabledProperty = DependencyProperty.RegisterAttached("IsEnabled", typeof(bool), typeof(ItemDragDropBehavior), new FrameworkPropertyMetadata(false, OnIsEnabledChanged));
        public static bool GetIsEnabled(DependencyObject target)
        {
            return (bool)target.GetValue(IsEnabledProperty);
        }
        public static void SetIsEnabled(DependencyObject target, bool value)
        {
            target.SetValue(IsEnabledProperty, value);
        }

        /// <summary>
        /// IsEnabled 添付プロパティ値変更イベントハンドラ
        /// </summary>
        /// <param name="d">イベント発行元</param>
        /// <param name="e">イベント引数</param>
        private static void OnIsEnabledChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var control = d as ItemsControl;
            // ItemDragDropBehavior ビヘイビアは ItemsControl 派生コントロールを対象としている
            if (control != null)
            {
                if (GetIsEnabled(control))
                {
                    // IsEnabled プロパティが true のときにイベントハンドラを登録する
                    control.PreviewMouseLeftButtonDown += control_PreviewMouseLeftButtonDown;
                    control.PreviewMouseMove += control_PreviewMouseMove;
                    control.PreviewDragEnter += control_PreviewDragEnter;
                    control.PreviewDragOver += control_PreviewDragOver;
                    control.PreviewDragLeave += control_PreviewDragLeave;
                    control.PreviewDrop += control_PreviewDrop;
                    control.PreviewMouseUp += control_PreviewMouseUp;
                }
                else
                {
                    // IsEnabled プロパティが false のときにイベントハンドラを登録解除する
                    control.PreviewMouseLeftButtonDown -= control_PreviewMouseLeftButtonDown;
                    control.PreviewMouseMove -= control_PreviewMouseMove;
                    control.PreviewDragEnter -= control_PreviewDragEnter;
                    control.PreviewDragOver -= control_PreviewDragOver;
                    control.PreviewDragLeave -= control_PreviewDragLeave;
                    control.PreviewDrop -= control_PreviewDrop;
                    control.PreviewMouseUp -= control_PreviewMouseUp;
                }
            }
        }
        #endregion IsEnabled 添付プロパティ

        #region private static フィールド
        /// <summary>
        /// マウス左ボタンによってこのビヘイビアによるドラッグ＆ドロップ操作シーケンスに移行したことを確認するためのフラグ
        /// </summary>
        private static bool _isMouseDown;

        /// <summary>
        /// ドラッグされているアイテムに対する論理親 (このビヘイビアが複数のコントロールに適用されることを考慮している)
        /// </summary>
        private static ItemsControl _parent;

        /// <summary>
        /// ドラッグされているアイテム
        /// </summary>
        private static object _draggedItem;

        /// <summary>
        /// ドラッグされているアイテムのインデックス
        /// </summary>
        private static int _draggedItemIndex;

        /// <summary>
        /// アイテムを掴むときのマウスポインタの位置
        /// </summary>
        private static Point _originPoint;
        #endregion private static フィールド

        #region イベントハンドラ
        private static void control_PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            _parent = sender as ItemsControl;
            var draggedItem = e.Source as FrameworkElement;
            if ((_parent == null) || (draggedItem == null))
                return;

            // ドラッグされるアイテムを取得する
            _draggedItem = _parent.ContainerFromElement(draggedItem);
            if (_draggedItem == null)
                return;

            // ドラッグされるアイテムのインデックスを取得する
            var itemsSourceList = _parent.ItemsSource as IList;
            itemsSourceList = itemsSourceList == null ? _parent.Items as IList : itemsSourceList;
            _draggedItemIndex = itemsSourceList.IndexOf(draggedItem);

            // ドラッグ開始位置を取得する
            var vw = Window.GetWindow(_parent);
            _originPoint = vw.PointToScreen(e.GetPosition(vw));

            _isMouseDown = true;
            Console.WriteLine("MouseLeftButtonDown");
        }

        private static void control_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (_isMouseDown && (e.LeftButton == MouseButtonState.Pressed))
            {
                // 現在位置を取得する
                var vw = Window.GetWindow(sender as DependencyObject);
                Point pt = vw.PointToScreen(e.GetPosition(vw));
                if (CheckDistance(_originPoint, pt))
                {
                    Console.WriteLine("DoDragDrop - Start");
                    DragDrop.DoDragDrop(sender as DependencyObject, _draggedItem, DragDropEffects.Move);
                    Console.WriteLine("DoDragDrop - end");

                    // ドラッグ＆ドロップが終わったらここに戻ってくるので
                    // ここで状態をクリアする
                    CleanUp();
                }
            }
        }

        private static void control_PreviewDragEnter(object sender, DragEventArgs e)
        {
            // Adorner の表示を復帰する
            Console.WriteLine("DragEnter : " + sender.ToString());
        }

        private static void control_PreviewDragOver(object sender, DragEventArgs e)
        {
            // ドラッグ操作中のマウス動作中は Adorner の位置を更新する
            Console.WriteLine("DragOver : " + sender.ToString());
        }

        private static void control_PreviewDragLeave(object sender, DragEventArgs e)
        {
            // ドラッグ操作中にコントロールからはみ出した場合は Adorner を隠す
            Console.WriteLine("DragLeave : " + sender.ToString());
        }

        private static void control_PreviewDrop(object sender, DragEventArgs e)
        {
            if (!_isMouseDown)
                return;

            var itemsControl = sender as ItemsControl;
            if (itemsControl == null)
                return;

            // ドロップする位置にあるアイテムのインデックスを取得する
            var itemsSourceList = itemsControl.ItemsSource as IList;
            itemsSourceList = itemsSourceList == null ? itemsControl.Items as IList : itemsSourceList;
            var dropTargetItem = e.Source as FrameworkElement;
            var dropTargetItemIndex = dropTargetItem != null ? itemsSourceList.IndexOf(dropTargetItem) : itemsSourceList.Count;
            dropTargetItemIndex = dropTargetItemIndex < 0 ? itemsSourceList.Count - 1 : dropTargetItemIndex;

            // ドラッグされたアイテムをドロップした位置に差し替える
            if (_parent == itemsControl)
            {
                itemsSourceList.Remove(_draggedItem);
                itemsSourceList.Insert(dropTargetItemIndex, _draggedItem);
            }
            else
            {
                var parentList = _parent.ItemsSource as IList;
                parentList = parentList == null ? _parent.Items as IList : parentList;
                parentList.Remove(_draggedItem);
                itemsSourceList.Insert(dropTargetItemIndex < 0 ? 0 : dropTargetItemIndex, _draggedItem);
            }

            e.Handled = true;
            Console.WriteLine("Drop");
        }

        private static void control_PreviewMouseUp(object sender, MouseButtonEventArgs e)
        {
            // 単純なクリックで Drop 操作されないようにここで状態をクリアする
            CleanUp();
            Console.WriteLine("MouseUp");
        }
        #endregion イベントハンドラ

        #region ヘルパ
        private static bool CheckDistance(Point x, Point y)
        {
            return true;

            // こっちにすると
            // アイテム同士の境界線ギリギリからドラッグを開始すると
            // 別のアイテムに IsMouseOver が移ってしまって
            // ドラッグするアイテムと選択中アイテムが一致しなくなってしまう
            //return (Math.Abs(x.X - y.X) >= SystemParameters.MinimumHorizontalDragDistance)
            //    || (Math.Abs(x.Y - y.Y) >= SystemParameters.MinimumVerticalDragDistance);
        }

        /// <summary>
        /// すべての状態をクリアする
        /// </summary>
        private static void CleanUp()
        {
            _isMouseDown = false;
            _parent = null;
            _draggedItem = null;
        }
        #endregion ヘルパ
    }
}
